import sys
import os
import pytest
from app import create_app

# Add the 'app' directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


@pytest.fixture()
def app():
    app = create_app()
    app.config.update(
        {
            "TESTING": True,
        }
    )

    # other setup can go here

    yield app

    # clean up / reset resources here


# can make client.get() and client.post() requests to the applications
@pytest.fixture()
def client(app):
    return app.test_client()



import os
import subprocess

from jsmin import jsmin

from autonomous import log


def search_files_with_extension(directory, extension, exclude=[]):
    """Recursively search for files with given extension in a directory tree."""
    matching_files = []
    for root, dirs, files in os.walk(directory):
        for e in exclude:
            if e in root:
                continue

        for file in files:
            for e in exclude:
                if e in file:
                    continue

            if file.endswith(extension):
                matching_files.append(os.path.join(root, file))
    return matching_files


def dartsass(path="static/style/main.scss", output="static/style/main.css", **kwargs):
    # print(f"==========================> dartsass  {path}, {output}, {kwargs}")
    subprocess.run(["sass", f"{path}:{output}"], capture_output=True)


def javascript(files="main", path="static/js", **kwargs):
    # Defining the path to the folder where the JS files are saved
    # Getting all the files from that folder
    if isinstance(files, list):
        for f in files:
            javascript(files=f, path=path, **kwargs)
    else:
        output = f"{path}/{files}.min.js"

        libfiles = search_files_with_extension(
            f"{path}/autojs", ".js", exclude=["tests/"]
        )
        lib_content = []
        for entry in libfiles:
            with open(entry, "r") as file:
                lib_content.append(file.read())

        # log(files, mainjs_content)

        # Create new master file
        with open(output, "w") as mainjs:
            # Add contents of files to master
            for i in lib_content:
                mainjs.write(f"{i}\n")
            js_file = open(f"{path}/{files}.js").read()
            mainjs.write(f"{js_file}\n")

            if kwargs.get("minified"):
                with open(mainjs, "r+") as js_file:
                    minified = jsmin(js_file.read())
                    js_file.seek(0)
                    js_file.write(minified)


def build_assets(
    csspath="static/style/main.scss",
    cssoutput="static/style/main.css",
    jsfiles="main",
    jspath="static/js",
):
    dartsass(path=csspath, output=cssoutput)
    javascript(files=jsfiles, path=jspath)



import pygit2
import os

class GHRemoteCallbacks(pygit2.RemoteCallbacks):
        def __init__(self):
            """
            _summary_
            """
            self.user = os.environ.get('GITHUB_USERNAME')
            self.token = os.environ.get('GITHUB_PAT')
            
        def credentials(self, url, username_from_url, allowed_types):
            """
            _summary_

            Args:
                url (_type_): _description_
                username_from_url (_type_): _description_
                allowed_types (_type_): _description_

            Returns:
                _type_: _description_
            """
            return pygit2.UserPass(self.user,self.token)
        
        def certificate_check(self, certificate, valid, host):
            """
            _summary_

            Args:
                certificate (_type_): _description_
                valid (_type_): _description_
                host (_type_): _description_

            Returns:
                _type_: _description_
            """
            return True



from ..db.autodb import Database


class ORM:
    _database = Database()

    def __init__(self, model):
        # breakpoint()
        table_name = model._table_name or model.__name__
        self.table = self._database.get_table(table=table_name, schema=model.attributes)
        self.name = self.table.name

    def save(self, data):
        return self.table.save(data)

    def get(self, pk):
        return self.table.get(pk)

    def all(self):
        return self.table.all()

    def search(self, **kwargs):
        return self.table.search(**kwargs)

    def find(self, **kwargs):
        return self.table.find(**kwargs)

    def delete(self, pk):
        return self.table.delete(pk)

    def flush_table(self):
        return self.table.clear()



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""
import os

import redis

from autonomous import log

from .table import Table


class Database:
    """
     _summary_

    _extended_summary_

    :return: _description_
    :rtype: _type_
    """

    port = os.getenv("REDIS_PORT", 6379)
    host = os.getenv("REDIS_HOST", "redis")
    password = os.getenv("REDIS_PASSWORD")
    username = os.getenv("REDIS_USERNAME")
    db = os.getenv("REDIS_DB")
    decode_responses = os.getenv("REDIS_DECODE")

    def __init__(self):
        """
        create an interface for your database
        """
        #log(self.username, self.password)
        options = {}

        if Database.username:
            options["username"] = Database.username
        if Database.password:
            options["password"] = Database.password
        if Database.decode_responses:
            options["decode_responses"] = Database.decode_responses

        self.db = redis.Redis(
            host=Database.host,
            port=Database.port,
            db=Database.db,
            **options,
        )
        self.tables = {}

    def get_table(self, table="default", schema=None):
        """
        opens the table from the file, which clears any changed data
        """
        if not self.db.exists(table):
            self.db.json().set(table, "$", {})

        self.tables[table] = Table(table, attributes=schema, db=self.db)
        return self.tables[table]



from autonomous import log
from autonomous.model.automodel import AutoModel


class Model(AutoModel):
    # set model default attributes
    attributes = {
        "name": "",
        "age": None,
    }



# opt : Optional[str]  # for optional attributes
# default : Optional[str] = "value"  # for default values
import copy
import pprint
from abc import ABC
from datetime import datetime

from autonomous import log

from .orm import ORM


class AutoModel(ABC):
    attributes = {}
    _table_name = ""
    _table = None
    _orm = ORM

    def __new__(cls, *args, **kwargs):
        """
        Create a new instance of the AutoModel.

        This method is responsible for creating a new instance of the AutoModel class.
        It sets default attributes, populates the object from the database if a primary key is provided,
        and handles additional keyword arguments.

        Args:
            cls: The class itself.
            *args: Positional arguments.
            **kwargs: Keyword arguments, including 'pk' for primary key.

        Returns:
            obj: The created AutoModel instance.
        """
        obj = super().__new__(cls)
        # set default attributes
        cls.attributes["pk"] = None
        cls.attributes["last_updated"] = datetime.now()
        # log(f"Creating {cls.__name__}")
        obj.pk = kwargs.pop("pk", None)
        result = cls.table().get(obj.pk) or {}
        for k, v in cls.attributes.items():
            setattr(obj, k, result.get(k, v))
        obj.__dict__ |= kwargs
        # log(obj, kwargs)
        cls._deserialize(obj.__dict__)
        return obj

    @classmethod
    def table(cls):
        if not cls._table or cls._table.name != cls.__name__:
            cls._table = cls._orm(cls)
        return cls._table

    @classmethod
    def model_name(cls):
        """
        Get the fully qualified name of this model.

        Returns:
            str: The fully qualified name of this model.
        """
        return f"{cls.__module__}.{cls.__name__}"

    def __repr__(self) -> str:
        """
        Return a string representation of the AutoModel instance.

        Returns:
            str: A string representation of the AutoModel instance.
        """
        return pprint.pformat(self.__dict__, indent=4, width=7, sort_dicts=True)

    def save(self):
        """
        Save this model to the database.

        Returns:
            int: The primary key (pk) of the saved model.
        """
        self.last_updated = datetime.now()
        result = self.serialize()
        record = {k: v for k, v in result.items() if k in self.attributes}
        self.pk = self.table().save(record)
        return self.pk

    @classmethod
    def get(cls, pk):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        if isinstance(pk, str) and pk.isdigit():
            pk = int(pk)

        result = cls.table().get(pk)
        return cls(**result) if result else None

    @classmethod
    def all(cls):
        """
        Get all models of this type.

        Returns:
            list: A list of AutoModel instances.
        """
        return [cls(**o) for o in cls.table().all()]

    @classmethod
    def search(cls, **kwargs):
        """
        Search for models containing the keyword values.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            list: A list of AutoModel instances that match the search criteria.
        """
        return [cls(**attribs) for attribs in cls.table().search(**kwargs)]

    @classmethod
    def find(cls, **kwargs):
        """
        Find the first model containing the keyword values and return it.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            AutoModel or None: The first matching AutoModel instance, or None if not found.
        """
        return cls.table().find(**kwargs)

    def delete(self):
        """
        Delete this model from the database.
        """
        self.table().delete(pk=self.pk)

    @classmethod
    def _serialize(self, val):
        if isinstance(val, list):
            for i, v in enumerate(val):
                val[i] = self._serialize(v)
        elif isinstance(val, dict):
            for k, v in val.items():
                val[k] = self._serialize(v)
        elif isinstance(val, datetime):
            val = {"_datetime": val.isoformat()}
        elif issubclass(val.__class__, AutoModel):
            val.save()
            val = {
                "_pk": val.pk,
                "_automodel": val.model_name(),
            }

        return val

    def serialize(self):
        """
        Serialize this model to a dictionary.

        Returns:
            dict: A dictionary representation of the serialized model.
        """
        return self._serialize(copy.deepcopy(self.__dict__))

    @classmethod
    def _deserialize(cls, val):
        if isinstance(val, dict):
            if "_automodel" in val:
                for model in AutoModel.__subclasses__():
                    if model.model_name() == val["_automodel"]:
                        val = model.get(val["_pk"])
                        break
            elif "_datetime" in val:
                val = datetime.fromisoformat(val["_datetime"])
            else:
                for k, v in val.items():
                    val[k] = cls._deserialize(v)
        elif isinstance(val, list):
            for i, v in enumerate(val):
                val[i] = cls._deserialize(v)
        return val

    @classmethod
    def deserialize(cls, vars):
        """
        Deserialize a dictionary to a model.

        Args:
            vars (dict): The dictionary to deserialize.

        Returns:
            AutoModel: A deserialized AutoModel instance.
        """
        return cls(**vars)



import os

import cloudinary.api
import cloudinary.uploader

from autonomous import log


class CloudinaryStorage:
    cloud_name = os.getenv("CLOUD_NAME")
    api_key = os.getenv("CLOUDINARY_KEY")
    api_secret = os.getenv("CLOUDINARY_SECRET")
    secure = True

    def __init__(self):
        cloudinary.config(
            cloud_name=self.cloud_name,
            api_key=self.api_key,
            api_secret=self.api_secret,
            secure=True,
        )

    @classmethod
    def get_metadata(cls, asset_id):
        # log(cls.api_secret, cls.api_key, cls.cloud_name)
        return cloudinary.api.resource_by_asset_id(asset_id)

    def search(self, name, **kwargs):
        query = name
        if folder := kwargs.get("folder"):
            query = f"{query} AND folder:{folder}"
        return cloudinary.Search().expression(query).execute()

    def update(self, key, **kwargs):
        metadata = self.get_metadata(key)
        return cloudinary.api.update(metadata["public_id"], **kwargs)

    def geturl(self, key):
        response = self.get_metadata(key)
        return response.get("url")

    def save(self, file, **kwargs):
        if folder := kwargs.get("folder"):
            try:
                cloudinary.api.subfolders(f"{folder}")
            except cloudinary.exceptions.NotFound as e:
                log(f"{e} -- Creating folder {folder}")
                cloudinary.api.create_folder(folder)
            finally:
                kwargs["asset_folder"] = folder

        try:
            response = cloudinary.uploader.upload(file, **kwargs)
        except Exception as e:
            log(f"Cloudinary Storage upload error: {response}")
            raise e

        return {"asset_id": response["asset_id"], "url": response["url"], "raw": None}

    def remove(self, asset_id, **kwargs):
        response = self.get_metadata(asset_id)
        cloudinary.uploader.destroy(response["public_id"], **kwargs)
